from nornir.core import Nornir
from nornir.core.inventory import ConnectionOptions

from ..utils import agador_cfg


def update_nornir_credentials(nr: Nornir):
    """
    Uses the credential map to update host credentials in this nornir instance
    """
    cred_map = agador_cfg("CRED_MAP")

    # set nornir default username and password
    nr.inventory.defaults.username = cred_map["defaults"]["username"]
    nr.inventory.defaults.password = cred_map["defaults"]["password"]
    if cred_map["defaults"]["enable"]:
        conn_options = ConnectionOptions(
            extras={"optional_args": {"secret": cred_map["defaults"]["enable"]}}
        )
        nr.inventory.defaults.connection_options["umnet_napalm"] = conn_options

    # for each host in the inventory, if there's a custom
    # match, set the credentials for that host
    for host in nr.inventory.hosts.values():
        # we only want to match a custom credential once for a particular host
        # the first match on the list
        found_custom_cred = False

        for custom_cred in cred_map["custom"]:
            if custom_cred["inventory_filter"](host) and not found_custom_cred:
                found_custom_cred = True
                host.username = custom_cred["username"]
                host.password = custom_cred["password"]

                if custom_cred.get("enable"):
                    conn_options = ConnectionOptions(
                        extras={"optional_args": {"secret": custom_cred["enable"]}}
                    )
                    host.connection_options["umnet_napalm"] = conn_options



from typing import Optional
from types import ModuleType

import importlib.machinery
import importlib.util

from os import getenv
from os.path import isdir
import re
import sys

import yaml
from decouple import UndefinedValueError, Config, RepositoryEnv
from cron_converter import Cron

from umnet_napalm.abstract_base import AbstractUMnetNapalm

from .mappers import credentials, save_to_db, save_to_file
from .utils import agador_cfg


class AgadorError(Exception):
    pass


class CommandMapError(Exception):
    def __init__(self, cmd: str, error_str: str):
        super().__init__(f"command map error for '{cmd}' - {error_str}")


class CredentialMapError(Exception):
    pass


def apply_config_settings(cfg_file: Optional[str]):
    """
    Get config settings from file, or if 'None' is provided, look for 'AGADOR_CFG'
    in environment. Update global variables with an 'AGADOR' attribute
    so we can reference the settings anywhere in the code.
    """

    cfg_file = cfg_file if cfg_file else getenv("AGADOR_CFG")
    if not cfg_file:
        raise AgadorError("No config file provided, and no AGADOR_CFG env set!")

    try:
        globals()["AGADOR"] = Config(RepositoryEnv(cfg_file)).repository.data

    except FileNotFoundError:
        raise AgadorError(f"Cannot locate agador config file {cfg_file}")

    load_command_map()
    load_credential_map()


def load_inventory_filters(file_name: str) -> ModuleType:
    """
    Loads inventory filters if they're not loaded already
    """

    # Prefer to only load the module once
    if "inventory_filters" not in sys.modules:
        loader = importlib.machinery.SourceFileLoader("inventory_filters", file_name)
        spec = importlib.util.spec_from_loader("inventory_filters", loader)
        inventory_filters = importlib.util.module_from_spec(spec)
        try:
            sys.modules["inventory_filters"] = inventory_filters
            loader.exec_module(inventory_filters)

        except Exception as e:
            raise AgadorError(f"Could not load inventory filters from {file_name}: {e}")

    # returning referece to modules for convenience
    return sys.modules["inventory_filters"]


def load_command_map():
    """
    Parses and validates command_map file, replacing text references
    to functions to references to the actual functions where applicable.

    Future task: pydantic-based validation
    """

    db_module = save_to_db
    file_module = save_to_file

    inventory_filters = load_inventory_filters(agador_cfg("INVENTORY_FILTERS"))

    with open(agador_cfg("CMD_MAP"), encoding="utf-8") as fh:
        cmd_map = yaml.safe_load(fh)

    output = {}

    ## agador netbox inventory filter
    globals()["AGADOR"]["NETBOX_INVENTORY_FILTER"] = cmd_map.get(
        "netbox_inventory_filter"
    )

    for cmd, data in cmd_map["commands"].items():
        output[cmd] = {}

        # validating frequency, which is required
        if "frequency" not in data:
            raise CommandMapError(cmd, "Must specify frequency")
        try:
            output[cmd]["frequency"] = Cron(data["frequency"])
        except ValueError:
            raise CommandMapError(cmd, "Invalid frequency - must be in crontab format")

        # validating umnet_napalm getter specification
        if "getter" not in data:
            raise CommandMapError(cmd, "Must specify umnet_napalm getter")
        if data["getter"] not in dir(AbstractUMnetNapalm):
            raise CommandMapError(cmd, f"Unknown umnet_napalm getter {data['getter']}")
        output[cmd]["getter"] = data["getter"]

        # validating and retrieving inventory filter fuction
        inv_filter = data.get("inventory_filter", None)
        if inv_filter:
            if inv_filter not in dir(inventory_filters):  # noqa: F821
                raise CommandMapError(cmd, f"Unknown inventory filter {inv_filter}")

            output[cmd]["inventory_filter"] = getattr(inventory_filters, inv_filter)  # noqa: F821

        # validating and retrieving save_to_file class
        file_data = data.get("save_to_file", None)
        if file_data:
            if "mapper" not in file_data:
                raise CommandMapError(cmd, "Must specify mapper for save_to_file")

            if "destination" not in file_data:
                raise CommandMapError(cmd, "Must specify destination for save_to_file")

            destination = resolve_envs(file_data["destination"])
            if not isdir(destination):
                raise CommandMapError(
                    cmd, f"Invalid desintation {destination} for save_to_file"
                )

            if file_data["mapper"] not in dir(file_module):
                raise CommandMapError(
                    cmd, f"Unknown save_to_file mapper {file_data['mapper']}"
                )

            output[cmd]["save_to_file"] = {
                "mapper": getattr(file_module, file_data["mapper"])(destination),
            }

        # validating and retrieving save_to_db class
        db_data = data.get("save_to_db", None)
        if db_data:
            if db_data not in dir(db_module):
                raise CommandMapError(cmd, f"Unknown save_to_db mapper {db_data}")

            output[cmd]["save_to_db"] = getattr(db_module, db_data)

        if not db_data and not file_data:
            raise CommandMapError(
                cmd, "Must specifiy either save_to_db or save_to_file"
            )

    globals()["AGADOR"]["CMD_MAP"] = output


def load_credential_map():
    """
    Parses and validates the credential map file, replacing references
    to filters and mappers with actual functions

    Future task: pydantic-based validation
    """

    required_default_fields = ["username", "password", "mapper", "enable"]
    required_custom_fields = ["mapper", "inventory_filter"]

    inventory_filters = load_inventory_filters(agador_cfg("INVENTORY_FILTERS"))

    with open(agador_cfg("CRED_MAP"), encoding="utf-8") as fh:
        cred_map = yaml.safe_load(fh)

    output = {}

    ##### default credential validation
    if "defaults" not in cred_map:
        raise CredentialMapError("Default credentials must be specified!")

    for reqd in required_default_fields:
        if reqd not in cred_map["defaults"]:
            raise CredentialMapError(f"default {reqd} is required")

    output["defaults"] = {}
    for k, v in cred_map["defaults"].items():
        if k == "mapper":
            if v not in dir(credentials):
                raise CredentialMapError(f"Invalid credential mapper {v}")
            output["defaults"]["mapper"] = getattr(credentials, v)
        else:
            output["defaults"][k] = v

    ##### custom credentials parsing
    output["custom"] = []
    seen_filters = []
    for cred in cred_map.get("custom", []):
        for reqd in required_custom_fields:
            if reqd not in cred:
                raise CredentialMapError(
                    f"Required custom credential {reqd} is missing"
                )

        if cred["inventory_filter"] not in dir(inventory_filters):
            raise CredentialMapError(
                f"Invalid inventory filter {cred['inventory_filter']}"
            )

        if cred["inventory_filter"] in seen_filters:
            raise CredentialMapError("Inventory filter values must be unique!")

        new_cred = {
            "inventory_filter": getattr(inventory_filters, cred["inventory_filter"]),
            "username": cred.get("username", output["defaults"]["username"]),
            "password": cred.get("password", output["defaults"]["password"]),
            "mapper": getattr(credentials, cred["mapper"]),
        }

        if "enable" in cred:
            new_cred["enable"] = cred["enable"]

        output["custom"].append(new_cred)
    globals()["AGADOR"]["CRED_MAP"] = resolve_credentials(output)


def resolve_envs(input_str: str) -> str:
    """
    Takes an input string and searches for all instances of '${ENV_VAR}', replacing
    ENV_VAR with the value in the .env file. Raises an exception
    if the ENV_VAR is not found
    """
    for m in re.finditer(r"\${(\w+)}", input_str):
        var = m.group(1)
        try:
            input_str = re.sub(r"\${" + var + "}", agador_cfg(var), input_str)
        except UndefinedValueError:
            raise ValueError(f"Invalid env var {m.group(1)} in {input_str}")

    return input_str


def resolve_credentials(cred_map):
    """
    Resolving all credentials in the cred map based on their
    mapper functions
    """

    # first we need to resolve all the credentials in the map
    _resolve_cred(cred_map["defaults"])
    for custom_cred in cred_map["custom"]:
        _resolve_cred(custom_cred)

    return cred_map


def _resolve_cred(cred):
    """
    Resolves enable and password fields in the credential map using the specified mapper
    """
    if "password" in cred:
        cred["password"] = cred["mapper"](cred["password"])

    if "enable" in cred:
        cred["enable"] = cred["mapper"](cred["enable"])



from typing import Optional

import logging
import threading
from datetime import datetime
import traceback

from nornir.core.task import AggregatedResult, Task, MultiResult
from nornir.core.inventory import Host

from .logging import LOG_FORMAT, ThreadLogFilter
from ..utils import send_email


logger = logging.getLogger(__name__)


class ProcessorBase:
    """
    convenience parent class so we don't have to define unused
    methods in child classes
    """

    def task_started(self, task: Task) -> None:
        pass

    def task_completed(self, task: Task, result: AggregatedResult) -> None:
        pass

    def task_instance_started(self, task: Task, host: Host) -> None:
        pass

    def task_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        pass

    def subtask_instance_started(self, task: Task, host: Host) -> None:
        pass

    def subtask_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        pass


class AgadorProcessor(ProcessorBase):
    """
    Agador processor. Designed to write to a specified logfile and optionally to stdout.
    If 'email_to' is not none, will send an email summarizing the results at the end.
    """

    def __init__(
        self,
        cmd_map: dict,
        total_hosts: int,
        logfile: str,
        email_from: str,
        email_to: Optional[str] = None,
        cli_output: Optional[bool] = False,
    ):
        self.cmd_map = cmd_map
        self.total_hosts = total_hosts
        self.email_from = email_from
        self.email_to = email_to
        self.logfile = logfile
        self.cli_output = cli_output
        self.start_time = datetime.now()

    def _log_output(self, output_str: str):
        with open(self.logfile, "a", encoding="utf-8") as fh:
            fh.write(output_str)

        if self.cli_output:
            print(output_str)

    def task_started(self, task: Task) -> None:
        msg = f"\n******* {task.name} Started at {self.start_time} *******\n"
        self._log_output(msg)

    def task_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        msg = f"** {host.name} "
        if result.failed:
            msg += f"failed at {datetime.now()}"
            for r in result:
                if r.exception:
                    msg += f" and raised the following exception: \n{r.exception}\n"

            msg += "**"
        else:
            msg += f"completed successfully at {datetime.now()}"

        msg += "**\n"
        self._log_output(msg)

    def task_completed(self, task: Task, result: AggregatedResult) -> None:
        """
        going to log, optionally print, and email the final results
        """

        # first step is to run any post-processing from the mappers in the command map
        for cmd, data in self.cmd_map.items():
            if data.get("save_to_file"):
                mapper = data["save_to_file"]["mapper"]
            elif data.get("save_to_db"):
                mapper = data["save_to_db"]

            try:
                if hasattr(mapper, "post_processing"):
                    getattr(mapper, "post_processing")()
            except Exception as e:
                logger.error(f"Error running post_procesing for {mapper} - {e}")
                logger.error(traceback.format_exc())

        # calculating how long the run took
        now = datetime.now()
        elapsed = (now - self.start_time).seconds
        m, s = divmod(elapsed, 60)
        h, m = divmod(m, 60)
        duration = f"{h} hours, {m} minutes, {s}, seconds"

        # generating outpupt message
        num_failed = len([r for r in result.values() if r.failed])
        num_passed = self.total_hosts - num_failed
        num_skipped = len(
            [r for r in result.values() if r.result == "No commands to execute"]
        )

        result_str = f"Completed run against {self.total_hosts} devices at {now}\n\n"
        result_str = f"Commands run: {', '.join(self.cmd_map)}\n"
        result_str += f"Elapsed time: {duration}\nTotal passed: {num_passed}\nTotal skipped: {num_skipped}\nTotal failed {num_failed}\n"

        if num_failed:
            result_str += "\nThe following devices failed:\n"
            for device, multi_results in result.items():
                # skip tasks that passed
                if not multi_results.failed:
                    continue

                errors = [str(r.result) for r in multi_results if r.failed]
                result_str += f"\t{device}: {','.join(errors)}\n"

        self._log_output(result_str)

        if self.email_to:
            send_email(
                email_from=self.email_from,
                email_to=self.email_to,
                subject=f"Run result at {now}",
                message=result_str,
            )


class TraceFile(ProcessorBase):
    """
    Class that sets up and tears down logging to a host-based tracefile
    """

    def __init__(self, trace_dir: str):
        self.trace_dir = trace_dir

    def task_instance_started(self, task: Task, host: Host) -> None:
        """
        Sets up tracefile and log handler filtering for the host's thread name
        """
        thread_name = threading.current_thread().name
        timestamp = datetime.now().strftime("%Y%d%m_%H%M%S")

        log_file = f"{self.trace_dir}/{host.name}_{timestamp}.trace"
        log_handler = logging.FileHandler(log_file)
        log_handler.addFilter(ThreadLogFilter(thread_name))
        log_handler.setFormatter(logging.Formatter(LOG_FORMAT))
        log_handler.name = host.name

        logger = logging.getLogger()
        logger.addHandler(log_handler)
        logger.setLevel(logging.DEBUG)

    def task_instance_completed(
        self, task: Task, host: Host, result: MultiResult
    ) -> None:
        """
        Tears down host log filter so that hosts using the same thread
        in the future don't get logged to this file
        """
        logger = logging.getLogger()

        # should only be one with this name
        [logger.removeHandler(h) for h in logger.handlers if h.name == task.host.name]



import logging
import argparse
import sys
import subprocess
from os import getenv

from datetime import datetime
import time

from .loaders import apply_config_settings
from .nornir.logging import LOG_FORMAT
from .utils import send_email, agador_cfg

logger = logging.getLogger("agador")

LOOP_INTERVAL = 10


def main():
    parser = argparse.ArgumentParser(description="Run agador")
    parser.add_argument(
        "--cfg-file",
        help="Configuration file.",
    )
    parser.add_argument(
        "-l", "--log-level", default="INFO", help="Set log level for agador only"
    )
    parser.add_argument("--echo", action="store_true", help="echo logfile to stdout")
    args = parser.parse_args()

    cfg_file = args.cfg_file if args.cfg_file else getenv("AGADOR_CFG")
    if not cfg_file:
        raise ValueError(
            "No config file provided as argument or env variable AGADOR_CFG"
        )

    apply_config_settings(cfg_file)

    cmd_map = agador_cfg("CMD_MAP")

    base_run_cmd = ["agador-run", "-l", args.log_level.upper(), "--cfg-file", cfg_file]

    # setting up logging
    file_handler = logging.handlers.RotatingFileHandler(
        agador_cfg("LOG_DIR") + "/agador.log", maxBytes=1024 * 1024 * 10, backupCount=20
    )
    file_handler.setFormatter(logging.Formatter(LOG_FORMAT))
    logger.addHandler(file_handler)
    logger.setLevel(args.log_level)
    if args.echo:
        stdout_handler = logging.StreamHandler(sys.stdout)
        stdout_handler.setFormatter(logging.Formatter(LOG_FORMAT))
        logger.addHandler(stdout_handler)
        base_run_cmd.append("--echo")

    email_to = agador_cfg("EMAIL_RECIPIENTS")
    if email_to:
        base_run_cmd.extend(["--email-result", email_to])

    current_run = None

    last_check = datetime.now()
    while True:
        # If there's a previous run and it has finished, check to make sure there
        # were no errors. If there were errors and we're configured to send emails,
        # send an email indicating what the error was.
        if current_run and current_run.poll():
            _, errs = current_run.communicate()
            if errs:
                error_msg = f"Agador run (PID {current_run.pid}) ended unexpectedly.\n\n{errs.decode('utf-8')}"
                logger.error(error_msg)
                if email_to:
                    send_email(
                        email_from=agador_cfg("EMAIL_ADDRESS"),
                        email_to=email_to,
                        subject="Unexpected error",
                        message=error_msg,
                    )

            current_run = None

        # waiting a bit, then checking the time
        time.sleep(LOOP_INTERVAL)
        now = datetime.now()

        # if any commands are scheduled to run between now and the last time
        # we checked the time, put them on a list of commands to run
        cmds_to_run = []
        for cmd, data in cmd_map.items():
            run_time = data["frequency"].schedule(now).prev()
            if run_time >= last_check:
                cmds_to_run.append(cmd)

        # If we have commands to run, first check to see if the previous run has completed.
        # if it has, run our new commands
        if cmds_to_run:
            if current_run and current_run.poll() is None:
                logger.error(
                    f"Agador wants to run cmds {','.join(cmds_to_run)} but previous run (PID {current_run.pid}) has not completed!"
                )
            else:
                cmd_args = base_run_cmd.copy()
                cmd_args.append("--cmds")
                cmd_args.extend(cmds_to_run)
                current_run = subprocess.Popen(cmd_args, stderr=subprocess.PIPE)
                logger.info(f"Starting agador run (PID {current_run.pid}): {cmd_args}")

        last_check = now


if __name__ == "__main__":
    main()



from ..credentials.cyberark import Cyberark
from ..utils import agador_cfg


def plaintext(cred: str) -> str:
    """
    Plain text mapper returns cred as is
    """
    return cred


def cyberark_umnet(cred: str) -> str:
    """
    Looks up a credential in the cyberark UMnet environment
    """
    c = Cyberark(env_file=agador_cfg("CYBERARK_ENV_FILE"), environment="UMNET")
    return c.query_cyberark(cred)


def cyberark_nso(cred: str) -> str:
    """
    Looks up a credential in the cyberark NSO environment
    """
    c = Cyberark(env_file=agador_cfg("CYBERARK_ENV_FILE"), environment="NSO")
    return c.query_cyberark(cred)



"""
Note - the Cyberark class below is adapted from one that Jeff Hagley wrote
and placed in umnet-scripts. Instead of augmenting that class, we've created
a new one here so that this package isn't dependent on the umnet-scripts
package, which has many dependencies of its own.
"""

import sys
from typing import Optional
from dataclasses import dataclass

from decouple import Config, RepositoryEnv
import requests
import urllib3

urllib3.disable_warnings()

CYBERARK_TIMEOUT = 30


class CyberarkLookupError(Exception):
    pass


class CyberarkTimeoutError(Exception):
    pass


class CyberarkAuthError(Exception):
    pass


@dataclass
class CyberarkEnvironment:
    name: str
    cert: tuple[str, str]
    appid: str
    safe: str


class Cyberark:
    """
    Class that interacts with the cyberark API.
    """

    def __init__(
        self, env_file: str = "cyberark.env", environment: Optional[str] = "UMNET"
    ):
        self.env_config = Config(RepositoryEnv(env_file))
        self.url_base = self.env_config.get("URL_BASE")
        self.url_aim_endpoint = self.env_config.get("URL_AIM_ENDPOINT")
        self.headers = {
            "Content-type": "application/json",
            "Accept": "application/json",
        }
        self.set_environment(environment)

    def set_environment(self, environment: str):
        environment = environment.upper()

        if environment == "UMNET":
            self.env = CyberarkEnvironment(
                environment,
                (self.env_config.get("CERT"), self.env_config.get("KEY")),
                self.env_config.get("APPID"),
                self.env_config.get("SAFE"),
            )
        elif environment == "NSO":
            self.env = CyberarkEnvironment(
                environment,
                (self.env_config.get("NSO_CERT"), self.env_config.get("NSO_KEY")),
                self.env_config.get("NSO_APPID"),
                self.env_config.get("NSO_SAFE"),
            )
        elif environment == "DEARBORN":
            self.env = CyberarkEnvironment(
                environment,
                (
                    self.env_config.get("DEARBORN_CERT"),
                    self.env_config.get("DEARBORN_KEY"),
                ),
                self.env_config.get("DEARBORN_APPID"),
                self.env_config.get("DEARBORN_SAFE"),
            )

        else:
            raise ValueError(f"Unknown cyberark environment {environment}")

    def query_cyberark(
        self, password_entity: str, environment: Optional[str] = None
    ) -> str:
        """
        Queries cyberark for a specific entity
        """
        if sys.version.startswith("3.8"):
            requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS = (  # pylint: disable=no-member
                "ALL:@SECLEVEL=1"
            )
        if environment:
            self.set_environment(environment)

        aim_url = (
            self.url_base
            + self.url_aim_endpoint
            + self.env.appid
            + self.env.safe
            + "&Object="
            + password_entity
        )
        try:
            acct = requests.get(
                aim_url,
                headers=self.headers,
                cert=self.env.cert,
                verify=True,
                timeout=CYBERARK_TIMEOUT,
            )
        except requests.exceptions.ReadTimeout:
            raise CyberarkTimeoutError(
                f"Cyberark query for {password_entity} timed out after {CYBERARK_TIMEOUT} seconds"
            )

        if acct.status_code == 404:
            raise CyberarkLookupError(
                f"No password found for {password_entity} environment {self.env.name}"
            )
        if acct.status_code == 403:
            raise CyberarkAuthError(f"Could not auth to cyberark for {self.env.name}")

        result = acct.json()

        if acct.status_code != 200 or not result or "Content" not in result:
            raise CyberarkLookupError(
                f"Invalid response from cyberark for {password_entity} environment {self.env.name}"
            )

        return result["Content"]

    def lookup_enable(self, host: str = "", environment: Optional[str] = None) -> str:
        """
        Looks up an 'enable' password for a specific host
        """
        if environment:
            self.set_environment(environment)

        return self.query_cyberark(f"{host}_enable")

    def lookup_username_password(
        self,
        username: str,
        host: Optional[str] = None,
        environment: Optional[str] = None,
    ) -> str:
        """
        Looks up the password for a specific username, optionally on a specific host.
        If there's no password found for that specific host, looks for that
        username for *all* hosts
        """
        if environment:
            self.set_environment(environment)

        # host passwords without usernames are stored in cyberark as 'name_user_name'
        if host is None:
            entity = f"{username}_user_{username}"
        else:
            entity = f"{host}_{username}"

        try:
            return self.query_cyberark(entity)
        except CyberarkLookupError:
            return self.query_cyberark(f"_{username}")



import logging
import traceback
from nornir.core.task import Result, Task
from nornir.core.exceptions import NornirSubTaskError
from sqlalchemy.engine import Engine
from sqlalchemy import create_engine
from umnet_napalm import get_network_driver
import re

from ..utils import get_device_cmd_list
from ..mappers.save_to_file import SaveResult
from ..mappers.save_to_db import ResultsToDb


logger = logging.getLogger(__name__)


def record_failure(result: Result, failure_message: str, e: Exception = None):
    """
    If a failure is detected, want to record a failed result
    with an appropriate message. If an exception was raised want to log that
    as well.
    """
    result.failed = True
    result.result = failure_message
    if e:
        result.exception = str(e) + "\n" + str(traceback.format_exc())


def process_device(
    task: Task,
    cmd_map: dict,
    db_url: str,
) -> Result:
    """
    Run a set of commands against a specific device based on the
    command map.
    """

    device_result = Result(host=task.host, result="Completed sucessfully")
    db_engine = create_engine(db_url)

    # first figure out which commands to run against this host
    cmd_list = get_device_cmd_list(cmd_map, task.host)

    # if no commands match for the host it's not a failure, and we want
    # to return before attempting to connect to the device.
    logger.debug(f"Running getters on {task.host.name}: {cmd_list}")
    if not cmd_list:
        logger.debug(f"No commands to execute for {task.host.name}")
        device_result.result = "No commands to execute"
        return device_result

    # finding that for multiprocessing it's cleaner to use the context
    # manager - by this we mean 'with network_driver as device' - because
    # connection failures that raise exceptions aren't otherwise well tolerated.
    # As a result we're not using the connection plugin and are instead establishing
    # the connection manually.
    try:
        network_driver = get_network_driver(task.host.platform)
    except NotImplementedError as e:
        record_failure(device_result, "Unsupported platform", e)
        return device_result

    parameters = {
        "hostname": task.host.hostname,
        "username": task.host.username,
        "password": task.host.password,
        "optional_args": {},
    }
    parameters.update(task.host.connection_options["umnet_napalm"].extras)

    getter_results = {}
    try:
        with network_driver(**parameters) as device:
            for cmd, getter in cmd_list.items():
                try:
                    method = getattr(device, getter)
                    getter_results[cmd] = method()

                except Exception as e:
                    # panos connection failures show up as 'command' failures
                    if type(e).__name__ == "PanXapiError" and re.search(
                        r"(service not known|Invalid Credential)", str(e)
                    ):
                        record_failure(
                            device_result, "Connection or Authentication Error", e
                        )
                        break
                    record_failure(device_result, f"Error executing {cmd}", e)

    # 'outer' failures are connection failures
    except Exception as e:
        record_failure(device_result, "Connection or Authentication Error", e)

    # for each command save data using the appropriate methods based
    # on the command map
    for cmd, result in getter_results.items():
        if "save_to_db" in cmd_map[cmd]:
            try:
                task.run(
                    name=f"{cmd}_save_to_db",
                    task=update_table,
                    result=result,
                    engine=db_engine,
                    mapper=cmd_map[cmd]["save_to_db"],
                )
            except NornirSubTaskError as e:
                record_failure(device_result, f"error saving {cmd} to db", e)

        if "save_to_file" in cmd_map[cmd]:
            try:
                task.run(
                    name=f"{cmd}_save_to_file",
                    task=save_to_file,
                    result=result,
                    mapper=cmd_map[cmd]["save_to_file"]["mapper"],
                )
            except NornirSubTaskError as e:
                record_failure(device_result, f"error saving {cmd} to file", e)

    return device_result


def update_table(
    task: Task,
    result: Result,
    engine: Engine,
    mapper: ResultsToDb,
) -> Result:
    """
    Generic task for taking the results from a previous "napalm_get"
    task and saving it to the database
    """
    mapper(task.host.name, task.host.hostname, result, engine)

    return Result(host=task.host)


def save_to_file(
    task: Task,
    result: dict,
    mapper: SaveResult,
) -> Result:
    """
    Task for saving to a file
    """
    mapper.write_to_file(task.host, result)

    return Result(host=task.host)



from typing import Optional

from nornir import InitNornir
from nornir.core import Nornir

from .connection_options import configure_connection_options
from .device_credentials import update_nornir_credentials

from ..utils import agador_cfg


def nornir_setup(
    device_filter: Optional[str] = None,
    role_filter: Optional[str] = None,
) -> Nornir:
    """
    Initializes Nornir to point at netbox, and to only care about active
    devices tied to a specific subset of device roles.
    Sets up logging. Populates default and custom passwords from cyberark. Returns
    customized Nornir instance
    """

    # Restrict what the netbox inventory plugin pulls if it was indicated
    # on the CLI
    filter_params = {}

    if device_filter:
        filter_params["name"] = device_filter
    elif role_filter:
        filter_params["role"] = [role_filter]

    # otherwise restrict based on agador netbox roles
    elif agador_cfg("NETBOX_INVENTORY_FILTER"):
        filter_params = agador_cfg("NETBOX_INVENTORY_FILTER")

    # Nornir initialization
    nr = InitNornir(
        runner={
            "plugin": "multiprocess",
            "options": {
                "num_workers": int(agador_cfg("NUM_WORKERS")),
            },
        },
        inventory={
            "plugin": "NetBoxInventory2",
            "options": {
                "nb_url": agador_cfg("NB_URL"),
                "nb_token": agador_cfg("NB_TOKEN"),
                "filter_parameters": filter_params,
                "ssl_verify": False,
            },
        },
        logging={
            "enabled": False,
        },
    )

    update_nornir_credentials(nr)
    configure_connection_options(nr)

    return nr



import sys
import re
import argparse
from datetime import datetime
import logging


from .nornir import nornir_setup
from .nornir.logging import configure_nornir_logging

from .nornir.tasks import process_device
from .nornir.processors import (
    AgadorProcessor,
    TraceFile,
)
from .utils import validate_email_list, AgadorError, agador_cfg
from .loaders import apply_config_settings


def main():
    parser = argparse.ArgumentParser(description="Run agador")

    parser.add_argument(
        "--cfg-file",
        help="Configuration file. You can also set this as AGADOR_CFG in your environennt",
    )

    filter_args = parser.add_mutually_exclusive_group()
    filter_args.add_argument(
        "--device", help="Restrict the update to a particular device"
    )
    filter_args.add_argument(
        "--role", help="Restrict the update to a specific netbox device role"
    )

    parser.add_argument(
        "--cmds",
        nargs="*",
        help="Limit update to a subset of tasks/commands",
    )
    parser.add_argument(
        "--email-result", help="Email result to a comma-separated list of addresses."
    )

    log_args = parser.add_mutually_exclusive_group()
    log_args.add_argument("-l", "--log-level", help="Set log level for agador only")
    log_args.add_argument("-L", "--LOG-LEVEL", help="set log level for all libraries")
    parser.add_argument(
        "--echo", action="store_true", help="echo logfile to stdout", default=True
    )
    parser.add_argument("--trace", action="store_true", help="Save device session logs")

    args = parser.parse_args()

    apply_config_settings(args.cfg_file)

    cmd_map = agador_cfg("CMD_MAP")

    # input validation
    if args.cmds:
        for cmd in args.cmds:
            if cmd not in cmd_map:
                print(
                    f"ERROR: {cmd} not in command map! valid commands: {','.join(cmd_map.keys())}"
                )
                sys.exit(1)

        # if we're filtering on what commands to run, remove them from the command map
        cmd_map = {k: v for k, v in cmd_map.items() if k in args.cmds}

    if args.trace and not args.device:
        sure = input(
            "Trace output is for debugging only. Are you sure you want to save session logs for ALL devices (y/n)? "
        )
        if re.match(r"[Yy]", sure):
            print(f"Fine. Logs will be saved at {agador_cfg('LOG_DIR')}")
        else:
            print("Good choice! Turning trace off.")
            args.trace = False

    if args.email_result:
        validate_email_list(args.email_result)

    # logging configuration
    if args.log_level:
        log_level = args.log_level
    elif args.LOG_LEVEL:
        log_level = args.LOG_LEVEL
    else:
        log_level = logging.INFO

    logfile = agador_cfg("LOG_DIR") + "/agador.log"
    logger = configure_nornir_logging(
        log_level,
        log_globally=bool(args.LOG_LEVEL),
        log_file=logfile,
        log_to_console=bool(args.echo),
    )

    # initialize nornir
    logger.info("initializing nornir")
    nr = nornir_setup(
        device_filter=args.device,
        role_filter=args.role,
    )

    logger.info(
        "Nornir initialization complete - inventory has %s items",
        len(nr.inventory.hosts.keys()),
    )

    if not (nr.inventory.hosts.keys()):
        logger.error("No matching hosts found in netbox inventory!")
        raise AgadorError("No matching hosts found in netbox inventory!")

    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    results_logfile = f"{agador_cfg('LOG_DIR')}/results_{timestamp}"

    # setting up processors
    processors = [
        AgadorProcessor(
            cmd_map=cmd_map,
            total_hosts=len(nr.inventory.hosts),
            email_from=agador_cfg("EMAIL_ADDRESS"),
            logfile=results_logfile,
            email_to=args.email_result,
            cli_output=args.echo,
        )
    ]

    if args.trace:
        processors.append(TraceFile(agador_cfg("LOG_DIR")))

    logger.debug("Starting run...")
    nr.with_processors(processors).run(
        task=process_device,
        cmd_map=cmd_map,
        db_url=agador_cfg("DB_URL"),
        on_failed=True,
    )


if __name__ == "__main__":
    main()



from typing import Optional, Dict, Any
import ipaddress
from email.message import EmailMessage
import smtplib
import re

from git import Repo

from nornir.core.inventory import Host


class AgadorError(Exception):
    pass


def is_ip_address(ip: str) -> bool:
    """
    Returns whether a string is an IP (eg 10.233.0.10 or fe80::1)
    """
    try:
        ipaddress.ip_address(ip)
    except ValueError:
        return False
    return True


def is_ip_or_prefix(ip: str) -> bool:
    """
    Returns whether a string is an ip (10.233.0.10) or
    IP + prefix (10.233.0.10/24). IPv6 works as well.
    """
    try:
        ipaddress.ip_interface(ip)
    except ValueError:
        return False
    return True


def agador_cfg(var: str) -> Any:
    """
    Looks up an agador config setting
    """
    try:
        from .loaders import AGADOR
    except ImportError:
        raise AgadorError(
            "Agador settings not loaded from file - run loaders/apply_config_settings first!"
        )

    return AGADOR.get(var, None)


def validate_email_list(input_str: str):
    """
    Makes sure that the input string is a comma-separated list
    of email addresses. Fun fact - the 'local' part of the email address
    can have a super wide range of characters, so we're just checking for non-whitespace
    The domain can have a-zA-Z0-9 and -, so we can be a bit more circumspect there.

    Function does nothing but throw an exception if the email is invalid
    """
    for email in re.split(r",", input_str):
        if not re.match(r"\S+@[\w\-.]+\.\w+$", email):
            raise ValueError(f"Invalid email {email}")


def git_update(repo_path: str, commit_message: str):
    """
    Updates git repo and if applicable, the remote origin.
    Will overwrite origin.
    """
    repo = Repo(repo_path)
    repo.git.status()

    repo.git.add(all=True)
    repo.index.commit(commit_message)
    try:
        origin = repo.remote(name="origin")
    except ValueError:
        origin = None

    if origin:
        origin.push()


def get_device_cmd_list(
    cmd_map: dict, host: Host, cmd_list_filter: Optional[list] = None
) -> Dict[str, str]:
    """
    Gets list of commands tied to a device based on its
    host inventory data. Optionally provide a list of commands to
    restrict the output to.

    Returns a dict mapping the agador command to the umnet_napalm getter
    """
    cmd_list = {}
    for cmd, data in cmd_map.items():
        if cmd_list_filter and cmd not in cmd_list_filter:
            continue

        if not data.get("inventory_filter") or data["inventory_filter"](host):
            cmd_list[cmd] = data["getter"]

    return cmd_list


def send_email(email_from: str, email_to: str, subject: str, message: str):
    """
    Sends an email via smtp.
    """
    msg = EmailMessage()
    msg["Subject"] = subject
    msg["From"] = email_from
    msg["To"] = email_to
    msg.set_content(message)

    with smtplib.SMTP("localhost") as s:
        s.send_message(msg)



import os
import re
import socket
from datetime import datetime
from git import Repo
from typing import Union
import logging
import pynetbox

from nornir.core.inventory import Host
from ..utils import agador_cfg

logger = logging.getLogger(__name__)


class SaveResult:
    """
    Generic "take nornir getter result and save it to a file" class
    """

    def __init__(self, destination: str):
        self._base_dir = destination

    def _get_hostname(self, netbox_name: str) -> str:
        """
        Removes stack member prefix from netbox hostnames
        """
        return re.sub(r"_\d$", "", netbox_name)

    def _set_output_file(self, host: Host) -> str:
        """
        Uses nornir host role data to determine sub-folder
        and ensures its created at the right path
        """
        role = host.data["device_role"]["slug"]
        if not os.path.exists(f"{self._base_dir}/{role}"):
            os.makedirs(f"{self._base_dir}/{role}")

        return f"{self._base_dir}/{role}/{self._get_hostname(host.name)}"

    def _git_update(self):
        """
        Updates git repo and if applicable, the remote origin.
        Will overwrite origin.
        """
        repo = Repo(self._base_dir)

        repo.git.add(all=True)

        if len(repo.index.diff("HEAD")) == 0:
            logger.info("No changes to repo, skipping commit")
            return

        formatted_date = datetime.strftime(datetime.now(), "%a %d %b %Y, %I:%M%p")
        commit_msg = f"Agador backup from {socket.gethostname()} at {formatted_date}"
        commit = repo.index.commit(commit_msg)
        logger.info(f"Committing {self._base_dir}: {commit}")

        try:
            origin = repo.remote(name="origin")
        except ValueError:
            logger.info("No origin found - skipping push")
            origin = None

        if origin:
            result = origin.push(force=True)
            urls = ",".join([u for u in origin.urls])
            summary = ",".join([r.summary for r in result])
            logger.info(f"Pushed to {urls}: {summary}")

    def write_to_file(self, host: Host, result: Union[dict, list]):
        """
        Saves data to file. Override in child classes if a
        different type of transformation needs to happen
        """
        if isinstance(result, dict):
            result = [result]
        with open(self._set_output_file(host), "w", encoding="utf-8") as fh:
            fh.write(",".join(result[0].keys()) + "\n")

            for row in result:
                values = [str(r) for r in row]
                fh.writelines(",".join(values) + "\n" for r in result)

    def post_processing(self):
        """
        Function that runs after all files are updated (post nornir run)
        """
        self._git_update()


class SaveConfig(SaveResult):
    """
    Saves running config from "get_config" to a file, redacting
    out passwords
    """

    def __init__(self, destination: str):
        super().__init__(destination)

        logger.debug(f"Checking for {self._base_dir}/obsolete")
        if not os.path.exists(f"{self._base_dir}/obsolete"):
            logger.debug("creating obsolete")
            os.makedirs(f"{self._base_dir}/obsolete")

    def write_to_file(self, host: Host, result: str):
        with open(self._set_output_file(host), "w", encoding="utf-8") as fh:
            for line in result.split("\n"):
                # ios - ignore ntp clock period
                if re.match(r"ntp clock-period", line):
                    continue

                # nxos - ignore current time statement
                if re.match(r"!Time:", line):
                    continue

                # iosxr date at top of file, format:
                # Tue May 13 00:00:34.586 EDT
                # Tue May  7 00:00:34.586 EDT
                if re.match(
                    r"[A-Z][a-z]{2} [A-Z][a-z]{2}\s{1,2}\d{1,2} \d{2}:\d{2}:\d{2}.\d{3} [A-Z]{3}$",
                    line,
                ):
                    continue

                # cisco type 7 password redaction
                if re.search(r"(password|key) 7", line):
                    line = "!!!" + re.sub(
                        r"(password|key) 7 \S+", r"\g<1> [redacted]", line
                    )

                # juniper type 9 password redaction
                if re.search(r"(secret|password) \"$9$", line):
                    line = "#" + re.sub(
                        r"(secret|password) \S+", r"\g<1> [redacted]", line
                    )

                fh.write(f"{line}\n")

    def post_processing(self):
        """
        Moves files for devices that are no longer listed in netbox as
        'active' to the 'retired' folder
        """
        logger.debug(f"Starting post processing for save_config in {self._base_dir}")

        roles = [
            r for r in os.listdir(self._base_dir) if not re.match(r"(\.|obsolete)", r)
        ]
        nb = pynetbox.api(agador_cfg("NB_URL"), token=agador_cfg("NB_TOKEN"))

        for role in roles:
            logger.debug(f"Processing role file {role}")

            role_dir = f"{self._base_dir}/{role}"
            nb_devices = [
                d.name for d in nb.dcim.devices.filter(role=role, status="active")
            ]

            for cfg_file in os.listdir(role_dir):
                if cfg_file not in nb_devices:
                    logger.debug(f"Moving {cfg_file} to obsolete directory")
                    os.rename(
                        f"{role_dir}/{cfg_file}",
                        f"{self._base_dir}/obsolete/{cfg_file}",
                    )

        super().post_processing()



import logging
import re

from nornir.core import Nornir
from nornir.core.inventory import ConnectionOptions

logger = logging.getLogger(__name__)

GLOBAL_TIMEOUT = 300

# maps firewalls to panorama instances. Note that the firewall name is assumed
# to be the name of the device in netbox
PANOS_FW_MAP = {
    "fw-dpss-1": "panorama-test.miserver.it.umich.edu",
    "fw-dpss-2": "panorama-test.miserver.it.umich.edu",
    "ngfw-1": "panorama-1.umnet.umich.edu",
    "ngfw-2": "panorama-1.umnet.umich.edu",
    "fw-umd-1": "panorama-1.umnet.umich.edu",
    "fw-umd-2": "panorama-1.umnet.umich.edu",
    "fw-umddc-1": "panorama-1.umnet.umich.edu",
    "fw-umddc-2": "panorama-1.umnet.umich.edu",
}


def configure_connection_options(nr: Nornir):
    """
    Applies custom connection options that make sense for our environment
    """

    for d in nr.inventory.hosts.values():
        if "umnet_napalm" not in d.connection_options:
            d.connection_options["umnet_napalm"] = ConnectionOptions(extras={})
        extras = d.connection_options["umnet_napalm"].extras

        # Setting our read timeouts. BIN routers and DNs take a long time so setting a very
        # large timeout for those
        if re.match(r"(bin-|d-\w+dn-)", d.name):
            extras["timeout"] = 7200
        else:
            extras["timeout"] = GLOBAL_TIMEOUT

        # for panorama-managed firewalls we needd to set the 'hostname' to be panorama,
        # the 'password' must be set as the api key, and the 'serial' must be saved as
        # an optional argument
        if d.name in PANOS_FW_MAP:
            optional_args = extras.get("optional_args", {})
            optional_args["api_key"] = d.password
            optional_args["serial"] = d.data["serial"]
            optional_args["panorama"] = PANOS_FW_MAP[d.name]
            extras["optional_args"] = optional_args
